from MeteorClient import MeteorClient
import uuid
import math
import base64
import os
import time
import requests


class Uploader():

    def __init__(self, client, collectionName, transport='ddp', callbacks={}, verbose=False):
        assert isinstance(client, MeteorClient)
        self.client = client
        self.collectionName = collectionName
        self.callbacks = callbacks
        self.verbose = verbose
        if not transport in ['ddp', 'http']:
            raise Exception('invalid transport.')
        self.transport = transport
        self.methodNames = {
            "_Abort": "_FilesCollectionAbort_" + collectionName,
            "_Write": "_FilesCollectionWrite_" + collectionName,
            "_Start": "_FilesCollectionStart_" + collectionName,
        }

    def _remove_callback(self, error, result):
        if error:
            print(error)
            return
        # print(result)

    def remove(self, selector, callback=None):
        self.client.call("_FilesCollectionRemove_" +
                         self.collectionName, selector, self._remove_callback)

    def _upload_write_callback(self, error, result):
        if error:
            self.error = True
            print(error)
            return
        # print(result)

    def _upload_end_callback(self, error, result):
        try:
            if self.callbacks.has_key('ended'):
                self.callbacks['ended'](error, result)
        except Exception as e:
            if self.verbose:
                print('error occured in ended callback')
        if error:
            self.error = True
            print(error)
            return
        self.finished = True
        if self.verbose:
            print('upload finished.')

    def _upload_start_callback(self, error, metaResult):
        try:
            if self.callbacks.has_key('started'):
                self.callbacks['started'](error, metaResult)
        except Exception as e:
            if self.verbose:
                print('error occured in started callback')

        if error:
            self.error = True
            print(error)
            return
        try:
            with open(self.filePath, "rb") as _file:
                for i in xrange(self.chunkCount):
                    try:
                        if self.callbacks.has_key('progress'):
                            self.callbacks['progress'](i, self.fileId)
                    except Exception as e:
                        if self.verbose:
                            print('error occured in progress callback')
                    if self.error:
                        raise
                    if self.verbose:
                        print('sending: ' + str(i + 1) +
                              '/' + str(self.chunkCount))
                    encoded_string = base64.b64encode(
                        _file.read(self.chunkSize))
                    if self.transport == 'ddp':
                        opts = {
                            "eof": False,
                            "fileId": self.fileId,
                            "binData": encoded_string,
                            "chunkId": i + 1,
                        }
                        self.client.call(self.methodNames['_Write'], [
                                         opts], self._upload_write_callback)
                    else:
                        baseurl = self.client.ddp_client.url
                        assert baseurl.startswith(
                            'ws://') and baseurl.endswith('/websocket')
                        uploadRoute = 'http' + \
                            baseurl[2:-10] + metaResult['uploadRoute']
                        headers = {
                            "x-eof": 0,
                            "x-fileid": self.fileId,
                            "x-chunkId": i + 1,
                            'content-type': 'text/plain'
                        }
                        r = requests.post(
                            uploadRoute, headers=headers, data=encoded_string)
                        r.raise_for_status()
        except Exception as e:
            print(e)
            self.error = True
            self.client.call(self.methodNames['_Abort'], [self.fileId])
        else:
            if self.verbose:
                print('sending EOF.')
            if self.transport == 'ddp':
                opts = {
                    "eof": True,
                    "fileId": self.fileId,
                }
                self.client.call(self.methodNames['_Write'], [
                                 opts], self._upload_end_callback)
            else:
                headers = {
                    "x-eof": 1,
                    "x-fileid": self.fileId,
                    'content-type': 'text/plain'
                }
                r = requests.post(uploadRoute, headers=headers, data='')
                r.raise_for_status()
                self.finished = True
                try:
                    if self.callbacks.has_key('ended'):
                        self.callbacks['ended'](r.status_code != requests.codes.ok, r)
                except Exception as e:
                    if self.verbose:
                        print('error occured in ended callback')
                if self.verbose:
                    print('upload finished.')

    def upload(self, filePath, chunkSize='dynamic', meta={}, fileType=None, fileId=None):
        self.filePath = filePath
        self.fileId = fileId or str(uuid.uuid4())
        fpath, fname = os.path.split(filePath)
        if fileType is None:
            try:
                import magic
                fileType = magic.from_file(filePath, mime=True)
            except Exception as e:
                import urllib
                import mimetypes
                url = urllib.pathname2url(filePath)
                fileType = mimetypes.guess_type(url)[0]
            if fileType is None:
                _, ext = os.path.splitext(fname)
                if ext != '':
                    fileType = 'application/' + ext
        assert fileType, 'unknown file type'

        statinfo = os.stat(filePath)
        fileSize = statinfo.st_size
        fileBase64Size = ((4 * fileSize / 3) + 3) & ~3
        if chunkSize == 'dynamic':
            chunkSize = fileBase64Size / 1000
        if chunkSize < 327680:
            chunkSize = 327680
        elif chunkSize > 1048576:
            chunkSize = 1048576
        chunkSize = int(math.floor(1.0 * chunkSize / 8) * 8)
        if self.transport == 'http':
            chunkSize = int(round(1.0 * chunkSize / 2))
        chunkCount = int(math.ceil(1.0 * fileBase64Size / chunkSize))

        self.fileType = fileType
        self.chunkCount = chunkCount
        self.chunkSize = chunkSize
        self.fileBase64Size = fileBase64Size
        self.fileSize = fileSize
        if self.verbose:
            print('file size: {}, chunk size: {}, chunk count: {}'.format(
                fileSize, chunkSize, chunkCount))
        error = False
        if self.verbose:
            print('start upload')
        opts = {
            "file": {"name": fname, "type": fileType, "size": fileSize, "meta": meta},
            "fileId": self.fileId,
            "chunkSize": chunkSize,
            "fileLength": 1 if chunkCount <= 0 else chunkCount,
        }
        self.finished = False
        self.error = False
        returnMeta = self.transport == 'http'
        self.client.call(self.methodNames['_Start'], [
                         opts, returnMeta], self._upload_start_callback)

if __name__ == '__main__':
    client = MeteorClient('ws://127.0.0.1:3000/websocket')
    client.connect()

    # upload example, work with Meteor-Files example: demo-simplest-upload
    # server code: https://github.com/VeliovGroup/Meteor-Files/tree/master/demo-simplest-upload
    client.subscribe('files');
    uploader = Uploader(client, 'files', transport='http', verbose=True)

    #import time
    #t0 = time.time()

    uploader.upload("test.jpeg", meta={"taskId":"JAqXA4yZFWnKh3G8x"})
    while not uploader.finished:
        time.sleep(0.1)

    #t1 = time.time()
    #print( 'time elapsed:%.1fs'%(t1-t0))



import keras
from keras import backend as K
import time
import os
import datetime
import numpy as np

class ProgressTracker(keras.callbacks.Callback):
    def __init__(self, task, save_weights=True):
        self.task = task
        self.save_weights = save_weights
        self.start_time = time.time()
        super(ProgressTracker, self).__init__()
        self.task.set('status.progress', 0)

    def on_batch_begin(self, batch, logs={}):
        if self.task.abort.is_set():
            self.task.set('status.error', 'interrupted')
            self.model.stop_training = True
            raise Exception('interrupted')

    def on_batch_end(self, batch, logs={}):
        info = ''
        for k, v in logs.items():
            if isinstance(v, (np.ndarray, np.generic) ):
                vstr = str(v.tolist())
            else:
                vstr = str(v)
            info += "{}: {}\n".format(k,vstr)
        self.elapsed_time = time.time()-self.start_time
        info += "elapsed_time: {:.2f}s".format(self.elapsed_time)
        self.task.update({
            'status.info': info,
            'status.progress': batch%100
        })

    def on_epoch_begin(self, epoch, logs={}):
        self.task.set('status.stage', 'epoch #'+str(epoch))
        if hasattr(self.model.optimizer, 'lr'):
            if self.task.get('config.learning_rate'):
                lr = float(self.task.get('config.learning_rate'))
                K.set_value(self.model.optimizer.lr, lr)
        else:
            self.task.set('status.error', 'Optimizer must have a "lr" attribute.')

    def on_epoch_end(self, epoch, logs={}):
        report = {}
        report['epoch'] = epoch
        for k, v in logs.items():
            report[k] = v
        if hasattr(self.model.optimizer, 'lr'):
            report['lr'] = K.get_value(self.model.optimizer.lr).tolist()
        self.task.push('output.training_history', report)
        self.task.set('output.elapsed_time', "%.2fs"%self.elapsed_time)
        self.task.set('output.last_epoch_update_time', datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))



import time

def rate_limited(period, damping=1.0, important=True):
  '''
  originally from https://github.com/tomasbasham/ratelimit

  Prevent a method from being called
  if it was previously called before
  a time widows has elapsed.
  :param period: The time window after which method invocations can continue.
  :param damping: A factor by which to dampen the time window.
  :return function: Decorated function that will forward method invocations if the time window has elapsed.
  '''
  frequency = damping / float(period)
  def decorate(func):
    last_called = [0.0]
    def func_wrapper(*args, **kargs):
      elapsed = time.clock() - last_called[0]
      left_to_wait = frequency - elapsed
      if left_to_wait > 0:
        if important:
            time.sleep(left_to_wait)
        else:
            return None
      ret = func(*args, **kargs)
      last_called[0] = time.clock()
      return ret
    return func_wrapper
  return decorate

class Metrics():

    def __init__(self, value, unit=''):
        self.value = value
        self.unit = unit
        self.__repr__ = self.__str__

    def __str__(self):
        return "{}{}".format(self.value, self.unit)

class NonBlockingStreamReader:

    def __init__(self, stream):
        '''
        stream: the stream to read from.
                Usually a process' stdout or stderr.
        '''
        self._s = stream
        self._q = Queue()

        def _populateQueue(stream, queue):
            '''
            Collect lines from 'stream' and put them in 'quque'.
            '''

            while True:
                line = stream.readline()
                if line:
                    queue.put(line)
                else:
                    self.end = True
                    break
                    #raise UnexpectedEndOfStream
                time.sleep(0.01)
        self.end = False
        self._t = threading.Thread(target=_populateQueue,
                                   args=(self._s, self._q))
        self._t.daemon = True
        self._t.start()  # start collecting lines from the stream

    def readline(self, timeout=None):
        try:
            return self._q.get(block=timeout is not None,
                               timeout=timeout)
        except Empty:
            if self.end:
                raise UnexpectedEndOfStream
            return None

class UnexpectedEndOfStream(Exception):
    pass



import logging
import threading
import os
import traceback
import time
import platform
import signal
from subprocess import Popen, PIPE, STDOUT
try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty  # python 3.x

class TaskProcessor(object):

    def __init__(self, task, widget, worker, save_widget_code=False):
        self.widget = widget
        self.task = task
        task.processor = self
        self.worker = worker
        self.meteorClient = self.worker.meteorClient
        assert not self.task is None
        self.isSubTask = bool(self.task.get('parent'))
        self.id = self.task.id
        self.parent = None if not task.get('parent') else task.get('parent')
        FORMAT = '%(asctime)-15s %(clientip)s %(user)-8s %(message)s'
        self.logger = logging.getLogger('taskProcessor')
        self.abort = self.task.abort
        self.exception = None
        self.running = False
        self.workdir = self.task.workdir
        self.make_workdir()
        self.changeCallbackDict = {}
        self.removeCallbackList = []
        self.requirements = {}
        self.resourcesOcuppied = {}
        if save_widget_code:
            self.save_widget_code()
        self.task.save()
        self.init()

    def save_widget_code(self, names=None):
        if self.widget.get('code_snippets'):
            self.make_workdir()
            if names is None:
                names = self.widget.get('code_snippets').keys()
            for k in names:
                k = k.replace('.', '_')
                code = self.widget.get('code_snippets')[k]
                with open(os.path.join(self.workdir, code['name']), 'w') as f:
                    f.write(code['content'])

    def init(self):
        pass

    def on_change(self, field, callback):
        if self.changeCallbackDict.has_key(field):
            self.changeCallbackDict[field].append(callback)
        else:
            self.changeCallbackDict[field] = [callback]

    def get_widget_code(self, name):
        name = name.replace('.', '_')
        if self.widget.get('code_snippets') and self.widget.get('code_snippets').has_key(name):
            return self.widget.get('code_snippets')[name]['content']
        else:
            return None

    def on_remove(self, callback):
        self.removeCallbackList.append(callback)

    def make_workdir(self):
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

    def update(self, key, value):
        try:
            if value is None and type(key) == type(dict()):
                self.meteorClient.call('tasks.update.worker', [self.id, self.worker.id, self.worker.token, {'$set': key}])
            else:
                self.meteorClient.call('tasks.update.worker', [self.id, self.worker.id, self.worker.token, {'$set': {key: value}}])
        except Exception as e:
            print('error ocurred during setting ' + key)

    def stop(self):
        self.abort.set()
        for subtask in self.task.subtasks:
            if subtask.processor:
                subtask.processor.abort.set()
        self.task.set('status.stage', 'aborting')
        print('stopping task...')

    def start(self, resources=None):
        self.abort.clear()
        self.running = True
        self.task.set('status.stage', 'starting')
        self.task.set('status.running', True)
        self.task.set('status.progress', -1)
        self.task.set('status.error', '')
        print('starting...')
        try:
            self.before()
            self.task.save()
            self.run(resources)
        except Exception as e:
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            traceback.print_exc()
            self.task.set('status.error', traceback.format_exc())
            self.end(force_quit=True)

    def end(self, force_quit=False):
        try:
            if not force_quit and len(self.task.subtasks) > 0:
                self.task.set('status.stage', 'finishing')
                lastfinished = 0
                while True:
                    finished = 0
                    total = len(self.task.subtasks)
                    for t in self.task.subtasks:
                        if not self.task.processor.running:
                            finished += 1
                        if self.abort.is_set():
                            break
                    if finished == total:
                        break
                    if self.abort.is_set():
                        break
                    if lastfinished != finished:
                        self.task.set('status.stage', '{}/{}'.format(finished,total))
                        lastfinished = finished
                    time.sleep(0.5)

                if self.abort.is_set():
                    self.task.set('status.stage', 'abort')
                else:
                    self.task.set('status.stage', 'done')
                    self.task.set('status.progress', 100)

            self.task.set('cmd', '')
            if self.task.get('status.progress') < 0:
                self.task.set('status.progress', 0)
            if 'ing' in self.task.get('status.stage'):
                self.task.set('status.stage', 'exited')
            self.task.save()
        except Exception as e:
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            traceback.print_exc()
            self.task.set('status.error', traceback.format_exc())
        finally:
            self.running = False
            self.task.set('status.running', False)
            self.task.set('visible2worker', False)
            try:
                self.after()
            except Exception as e:
                print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
                traceback.print_exc()

    def before(self):
        pass
        #self.task.set('output', {})

    def after(self):
        pass

    def after_runtime_error(self):
        print('after_runtime_error')

    def task_arguments(self, resources, env):
        return []

    def name(self):
        return self.task.get('name')

    def run(self, resources=None):
        '''
        should call self.end() after all process is over
        '''
        self.task.set('status.info', 'run is not implemented')
        self.end()

    def check_subtasks(self):
        status = {}
        prog = 0
        finished = 0
        for subtask in self.task.subtasks:
            prog +=subtask.get('status.progress')
            if not subtask.processor.running:
                finished +=1
        status['progress'] = prog
        status['finished'] = finished
        status['total'] = len(self.task.subtasks)
        return status


class ThreadedTaskProcessor(TaskProcessor):
    def __init__(self, task, widget, worker, process=None, **kwargs):
        self.process = process
        super(ThreadedTaskProcessor, self).__init__(task, widget, worker, **kwargs)

    def task_arguments(self, resources, env):
        return []

    def run(self, resources=None):
        self.taskThread = threading.Thread(target=self.run_thread,
                                           args=self.task_arguments(resources, None))
        self.taskThread.daemon = True
        self.taskThread.start()

    def run_thread(self, *args):
        try:
            self.process_task(*args)
            if self.process:
                self.process(self.task, *args)
        except Exception as e:
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            traceback.print_exc()
            self.task.set('status.error', traceback.format_exc())
        finally:
            self.end()

    def process_task(self, *args):
        pass


class ProcessTaskProcessor_(TaskProcessor):

    def process_output(self, line):
        self.task.set('status.progress', int(line))
        print('processing line: ' + line)
        sys.stdout.flush()
        return True

    def task_arguments(self, resources, env):
        return ['python', 'dummytask.py']

    def periodic_check(self, process):
        pass

    def run(self, resources=None):
        env = os.environ.copy()
        args = self.task_arguments(resources, env)
        if not args:
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            self.logger.error('Could not create the arguments for Popen')
            self.task.set('status.error', 'Could not create the arguments for Popen')
            return False
        # Convert them all to strings
        args = [str(x) for x in args if str(x) != '']
        self.logger.info('%s task started.' % self.name())
        unrecognized_output = []
        import sys
        env['PYTHONPATH'] = os.pathsep.join(
            ['.', self.workdir, env.get('PYTHONPATH', '')] + sys.path)
        # https://docs.python.org/2/library/subprocess.html#converting-argument-sequence
        if platform.system() == 'Windows':
            args = ' '.join(args)
        self.logger.info('Task subprocess args: {}'.format(args))
        try:
            p = Popen(args, bufsize=0, stdout=PIPE, stderr=STDOUT,
                      shell=False, universal_newlines=True)
        except Exception as e:
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            traceback.print_exc()
            self.task.set('status.error', traceback.format_exc())
            self.end(force_quit=True)
            return False
        # run the shell as a subprocess:

        nbsr = NonBlockingStreamReader(p.stdout)
        try:
            sigterm_time = None  # When was the SIGTERM signal sent
            sigterm_timeout = 2  # When should the SIGKILL signal be sent
            # get the output
            endofstream = False
            while p.poll() is None or not endofstream:
                try:
                    line = nbsr.readline(0.1)
                except(UnexpectedEndOfStream):
                    endofstream = True
                try:
                    self.periodic_check(p)
                except:
                    traceback.print_exc()

                if line is not None:
                    # Remove whitespace
                    line = line.strip()
                if line:
                    try:
                        if not self.process_output(line):
                            self.logger.warning('%s unrecognized output: %s' % (
                                self.name(), line.strip()))
                            unrecognized_output.append(line)
                    except:
                        print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
                        traceback.print_exc()
                        self.task.set('status.error', traceback.format_exc())
                else:
                    time.sleep(0.05)

                if self.abort.is_set():
                    if sigterm_time is None:
                        # Attempt graceful shutdown
                        p.send_signal(signal.SIGTERM)
                        sigterm_time = time.time()
                if sigterm_time is not None and (time.time() - sigterm_time > sigterm_timeout):
                    p.send_signal(signal.SIGKILL)
                    self.logger.warning(
                        'Sent SIGKILL to task "%s"' % self.name())
                    time.sleep(0.1)
        except:
            traceback.print_exc()
            try:
                p.terminate()
            except Exception as e:
                print('error occured during terminating a process.')
            raise
        if self.abort.is_set():
            self.end(force_quit=True)
            return False
        elif p.returncode != 0:
            # Report that this task is finished
            self.logger.error('%s task failed with error code %s' %
                              (self.name(), str(p.returncode)))
            if self.exception is None:
                self.exception = 'error code %s' % str(p.returncode)
                if unrecognized_output:
                    if self.traceback is None:
                        self.traceback = '\n'.join(unrecognized_output)
                    else:
                        self.traceback = self.traceback + \
                            ('\n'.join(unrecognized_output))
            self.after_runtime_error()
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            self.task.set('status.error', '%s task failed with error code %s' % (self.name(), str(p.returncode)))
            self.end(force_quit=True)
            return False
        else:
            self.task.set('status.progress', 100)
            self.end(force_quit=True)
            self.logger.info('%s task completed.' % self.name())
            return True

class ProcessTaskProcessor(ProcessTaskProcessor_):
    def run(self, resources=None):
        self.taskThread = threading.Thread(target=self.run_thread,
                                           args=[resources])
        self.taskThread.daemon = True
        self.taskThread.start()

    def run_thread(self, *args):
        try:
            super(ProcessTaskProcessor, self).run(*args)
        except Exception as e:
            print('error from task, taskName:{} taskId:{} widgetId:{}'.format(self.task.get('name'), self.task.id, self.task.get('widgetId')))
            traceback.print_exc()
            self.task.set('status.error', traceback.format_exc())
        finally:
            self.end()

class NonBlockingStreamReader:

    def __init__(self, stream):
        '''
        stream: the stream to read from.
                Usually a process' stdout or stderr.
        '''
        self._s = stream
        self._q = Queue()

        def _populateQueue(stream, queue):
            '''
            Collect lines from 'stream' and put them in 'quque'.
            '''

            while True:
                line = stream.readline()
                if line:
                    queue.put(line)
                else:
                    self.end = True
                    break
                    #raise UnexpectedEndOfStream
                time.sleep(0.01)
        self.end = False
        self._t = threading.Thread(target=_populateQueue,
                                   args=(self._s, self._q))
        self._t.daemon = True
        self._t.start()  # start collecting lines from the stream

    def readline(self, timeout=None):
        try:
            return self._q.get(block=timeout is not None,
                               timeout=timeout)
        except Empty:
            if self.end:
                raise UnexpectedEndOfStream
            return None


class UnexpectedEndOfStream(Exception):
    pass



import time
import datetime

from DDPClient import DDPClient

from pyee import EventEmitter
import hashlib

class MeteorClientException(Exception):
    """Custom Exception"""
    pass

class CollectionData(object):
    def __init__(self):
        self.data = {}

    def add_data(self, collection, id, fields):
        if collection not in self.data:
            self.data[collection] = {}
        if not id in self.data[collection]:
            self.data[collection][id] = {}
        for key, value in fields.items():
            self.data[collection][id][key] = value

    def change_data(self, collection, id, fields, cleared):
        for key, value in fields.items():
            self.data[collection][id][key] = value

        for key in cleared:
            del self.data[collection][id][key]

    def remove_data(self, collection, id):
        del self.data[collection][id]

class MeteorClient(EventEmitter):
    def __init__(self, url, auto_reconnect=True, auto_reconnect_timeout=0.5, debug=False):
        EventEmitter.__init__(self)
        self.collection_data = CollectionData()
        self.ddp_client = DDPClient(url, auto_reconnect=auto_reconnect,
                                    auto_reconnect_timeout=auto_reconnect_timeout, debug=debug)
        self.ddp_client.on('connected', self.connected)
        self.ddp_client.on('socket_closed', self.closed)
        self.ddp_client.on('failed', self.failed)
        self.ddp_client.on('added', self.added)
        self.ddp_client.on('changed', self.changed)
        self.ddp_client.on('removed', self.removed)
        self.ddp_client.on('reconnected', self._reconnected)
        self.connected = False
        self.subscriptions = {}
        self._login_data = None
        self._login_token = None

    def connect(self):
        """Connect to the meteor server"""
        self.ddp_client.connect()

    def close(self):
        """Close connection with meteor server"""
        self.ddp_client.close()

    def _reconnected(self):
        """Reconnect

        Currently we get a new session every time so we have
        to clear all the data an resubscribe"""

        if self._login_data or self._login_token:

            def reconnect_login_callback(error, result):
                if error:
                    if self._login_token:
                        self._login_token = None
                        self._login(self._login_data,
                                    callback=reconnect_login_callback)
                        return
                    else:
                        raise MeteorClientException(
                            'Failed to re-authenticate during reconnect')

                self.connected = True
                self._resubscribe()

            if self._login_token:
                self._resume(self._login_token,
                             callback=reconnect_login_callback)
            else:
                self._login(self._login_data,
                            callback=reconnect_login_callback)
        else:
            self.connected = True
            self._resubscribe()

    def _resubscribe(self):
        self.collection_data.data = {}
        cur_subs = self.subscriptions.items()
        self.subscriptions = {}
        for name, value in cur_subs:
            self.subscribe(name, value['params'])
        self.emit('reconnected')
    #
    # Account Management
    #

    def login(self, user, password, token=None, callback=None):
        """Login with a username and password

        Arguments:
        user - username or email address
        password - the password for the account

        Keyword Arguments:
        token - meteor resume token
        callback - callback function containing error as first argument and login data"""
        # TODO: keep the tokenExpires around so we know the next time
        #       we need to authenticate

        # hash the password
        hashed = hashlib.sha256(password).hexdigest()
        # handle username or email address
        if '@' in user:
            user_object = {
                'email': user
            }
        else:
            user_object = {
                'username': user
            }
        password_object = {
            'algorithm': 'sha-256',
            'digest': hashed
        }

        self._login_token = token
        self._login_data = {'user': user_object, 'password': password_object}

        if token:
            self._resume(token, callback=callback)
        else:
            self._login(self._login_data, callback=callback)

    def _resume(self, token, callback=None):
        login_data = {'resume': token}
        self._login(login_data, callback=callback)

    def _login(self, login_data, callback=None):
        self.emit('logging_in')

        def logged_in(error, data):
            if error:
                if self._login_token:
                    self._login_token = None
                    self._login(self._login_data, callback=callback)
                    return
                if callback:
                    callback(error, None)
                return

            self._login_token = data['token']

            if callback:
                callback(None, data)
            self.emit('logged_in', data)

        self.ddp_client.call('login', [login_data], callback=logged_in)

    def logout(self, callback=None):
        """Logout a user

        Keyword Arguments:
        callback - callback function called when the user has been logged out"""
        self.ddp_client.call('logout', [], callback=callback)
        self.emit('logged_out')

    #
    # Meteor Method Call
    #

    def call(self, method, params, callback=None):
        """Call a remote method

        Arguments:
        method - remote method name
        params - remote method parameters

        Keyword Arguments:
        callback - callback function containing return data"""
        self._wait_for_connect()
        self.ddp_client.call(method, params, callback=callback)

    #
    # Subscription Management
    #

    def subscribe(self, name, params=[], callback=None):
        """Subscribe to a collection
        Arguments:
        name - the name of the publication
        params - the subscription parameters

        Keyword Arguments:
        callback - a function callback that returns an error (if exists)"""
        self._wait_for_connect()

        def subscribed(error, sub_id):
            if error:
                self._remove_sub_by_id(sub_id)
                if callback:
                    callback(error.get('reason'))
                return
            if callback:
                callback(None)
            self.emit('subscribed', name)

        if name in self.subscriptions:
            raise MeteorClientException('Already subcribed to {}'.format(name))

        sub_id = self.ddp_client.subscribe(name, params, subscribed)
        self.subscriptions[name] = {
            'id': sub_id,
            'params': params
        }

    def unsubscribe(self, name):
        """Unsubscribe from a collection

        Arguments:
        name - the name of the publication"""
        self._wait_for_connect()
        if name not in self.subscriptions:
            raise MeteorClientException('No subscription for {}'.format(name))
        self.ddp_client.unsubscribe(self.subscriptions[name]['id'])
        del self.subscriptions[name]
        self.emit('unsubscribed', name)

    #
    # Collection Management
    #

    def find(self, collection, selector={}):
        """Find data in a collection

        Arguments:
        collection - collection to search

        Keyword Arguments:
        selector - the query (default returns all items in a collection)"""
        results = []
        for _id, doc in self.collection_data.data.get(collection, {}).items():
            doc.update({'_id': _id})
            if selector == {}:
                results.append(doc)
            for key, value in selector.items():
                if key in doc and doc[key] == value:
                    results.append(doc)
        return results

    def find_one(self, collection, selector={}):
        """Return one item from a collection

        Arguments:
        collection - collection to search

        Keyword Arguments:
        selector - the query (default returns first item found)"""
        for _id, doc in self.collection_data.data.get(collection, {}).items():
            doc.update({'_id': _id})
            if selector == {}:
                return doc
            for key, value in selector.items():
                if key in doc and doc[key] == value:
                    return doc
        return None

    def insert(self, collection, doc, callback=None):
        """Insert an item into a collection

        Arguments:
        collection - the collection to be modified
        doc - The document to insert. May not yet have an _id attribute,
        in which case Meteor will generate one for you.

        Keyword Arguments:
        callback - Optional. If present, called with an error object as the first argument and,
        if no error, the _id as the second."""
        self.call("/" + collection + "/insert", [doc], callback=callback)

    def update(self, collection, selector, modifier, callback=None):
        """Insert an item into a collection

        Arguments:
        collection - the collection to be modified
        selector - specifies which documents to modify
        modifier - Specifies how to modify the documents

        Keyword Arguments:
        callback - Optional. If present, called with an error object as the first argument and,
        if no error, the number of affected documents as the second."""
        self.call("/" + collection + "/update", [selector, modifier], callback=callback)

    def remove(self, collection, selector, callback=None):
        """Remove an item from a collection

        Arguments:
        collection - the collection to be modified
        selector - Specifies which documents to remove

        Keyword Arguments:
        callback - Optional. If present, called with an error object as its argument."""
        self.call("/" + collection + "/remove", [selector], callback=callback)

    #
    # Event Handlers
    #

    def connected(self):
        self.connected = True
        self.emit('connected')

    def closed(self, code, reason):
        self.connected = False
        self.emit('closed', code, reason)

    def failed(self, data):
        self.emit('failed', str(data))

    def added(self, collection, id, fields):
        self.collection_data.add_data(collection, id, fields)
        self.emit('added', collection, id, fields)

    def changed(self, collection, id, fields, cleared):
        self.collection_data.change_data(collection, id, fields, cleared)
        self.emit('changed', collection, id, fields, cleared)

    def removed(self, collection, id):
        self.collection_data.remove_data(collection, id)
        self.emit('removed', collection, id)

    #
    # Helper functions
    #

    def _time_from_start(self, start):
        now = datetime.datetime.now()
        return now - start

    def _remove_sub_by_id(self, sub_id):
        for name, cur_sub_id in self.subscriptions.items():
            if cur_sub_id == sub_id:
                del self.subscriptions[name]

    def _wait_for_connect(self):
        start = datetime.datetime.now()
        while not self.connected and self._time_from_start(start).seconds < 5:
            time.sleep(0.1)

        if not self.connected:
            raise MeteorClientException('Could not subscribe because a connection has not been established')
